@@ -36,10 +36,10 @@
 
 #include <gio/gio.h>
 #include <gio/gunixfdlist.h>
-#include "glib-backports.h"
 #include "document-portal-dbus.h"
 #include "document-store.h"
 #include "src/xdp-app-info.h"
+#include "src/xdp-app-info-registry.h"
 #include "src/xdp-utils.h"
 #include "permission-db.h"
 #include "permission-store-dbus.h"
@@ -69,6 +69,8 @@ static dev_t fuse_dev = 0;
 static GQueue get_mount_point_invocations = G_QUEUE_INIT;
 static XdpDbusDocuments *dbus_api;
 
+XdpAppInfoRegistry *app_info_registry;
+
 G_LOCK_DEFINE (db);
 
 char **
@@ -650,10 +652,10 @@ app_has_file_access (const char *target_app_id,
       res = xdp_spawn (&error, "flatpak", "info", arg, target_app_id, NULL);
     }
 
-  g_strchomp (res);
-
   if (res)
     {
+      g_strchomp (res);
+
       if (strcmp (res, "read-write") == 0)
         return TRUE;
 
@@ -1178,7 +1180,10 @@ handle_method (GCallback              method_callback,
   g_autoptr(XdpAppInfo) app_info = NULL;
   PortalMethod portal_method = (PortalMethod)method_callback;
 
-  app_info = xdp_invocation_ensure_app_info_sync (invocation, NULL, &error);
+  app_info = xdp_app_info_registry_ensure_for_invocation_sync (app_info_registry,
+                                                               invocation,
+                                                               NULL,
+                                                               &error);
   if (app_info == NULL)
     g_dbus_method_invocation_return_gerror (invocation, error);
   else
@@ -1496,7 +1501,8 @@ portal_get_host_paths (GDBusMethodInvocation *invocation,
 }
 
 static void
-peer_died_cb (const char *name)
+on_peer_disconnect (const char *name,
+                    gpointer    user_data)
 {
   stop_file_transfers_for_sender (name);
 }
@@ -1511,6 +1517,8 @@ on_bus_acquired (GDBusConnection *connection,
 
   dbus_api = xdp_dbus_documents_skeleton_new ();
 
+  app_info_registry = xdp_app_info_registry_new ();
+
   xdp_dbus_documents_set_version (XDP_DBUS_DOCUMENTS (dbus_api), 5);
 
   g_signal_connect_swapped (dbus_api, "handle-get-mount-point", G_CALLBACK (handle_get_mount_point), NULL);
@@ -1530,7 +1538,7 @@ on_bus_acquired (GDBusConnection *connection,
   g_dbus_interface_skeleton_set_flags (file_transfer,
                                        G_DBUS_INTERFACE_SKELETON_FLAGS_HANDLE_METHOD_INVOCATIONS_IN_THREAD);
 
-  xdp_connection_track_name_owners (connection, peer_died_cb);
+  xdp_connection_track_peer_disconnect (connection, on_peer_disconnect, NULL);
 
   if (!g_dbus_interface_skeleton_export (G_DBUS_INTERFACE_SKELETON (dbus_api),
                                          connection,
@@ -1723,7 +1731,7 @@ main (int    argc,
 
   g_autoptr(GError) error = NULL;
   g_autofree char *path = NULL;
-  GDBusConnection *session_bus;
+  g_autoptr(GDBusConnection) session_bus = NULL;
   g_autoptr(GOptionContext) context = NULL;
   GDBusMethodInvocation *invocation;
 
