@@ -25,6 +25,7 @@
 #include "config.h"
 
 #include <locale.h>
+#include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -40,6 +41,7 @@
 
 #include "file-transfer.h"
 #include "src/xdp-app-info.h"
+#include "src/xdp-app-info-registry.h"
 #include "src/xdp-utils.h"
 #include "document-portal-dbus.h"
 #include "document-enums.h"
@@ -48,6 +50,8 @@
 
 static XdpDbusFileTransfer *file_transfer;
 
+extern XdpAppInfoRegistry *app_info_registry;
+
 typedef struct
 {
   char *path;
@@ -148,11 +152,12 @@ lookup_transfer (const char *key)
 
   G_LOCK (transfers);
   transfer = (FileTransfer *)g_hash_table_lookup (transfers, key);
-  if (transfer)
-    g_object_ref (transfer);
   G_UNLOCK (transfers);
 
-  return transfer;
+  if (!transfer)
+    return NULL;
+
+  return g_object_ref (transfer);
 }
 
 static FileTransfer *
@@ -172,13 +177,14 @@ file_transfer_start (XdpAppInfo *app_info,
   transfer->files = g_ptr_array_new_with_free_func (exported_file_free);
 
   G_LOCK (transfers);
-  do {
-    guint64 key;
-    g_free (transfer->key);
-    key = g_random_int ();
-    key = (key << 32) | g_random_int ();
-    transfer->key = g_strdup_printf ("%" G_GUINT64_FORMAT, key);
-  }
+  do
+    {
+      uint64_t key;
+      g_free (transfer->key);
+      key = g_random_int ();
+      key = (key << 32) | g_random_int ();
+      transfer->key = g_strdup_printf ("%" G_GUINT64_FORMAT, key);
+    }
   while (g_hash_table_contains (transfers, transfer->key));
   g_hash_table_insert (transfers, transfer->key, g_object_ref (transfer));
   G_UNLOCK (transfers);
@@ -227,9 +233,9 @@ file_transfer_stop (FileTransfer *transfer)
 
 static void
 file_transfer_add_file (FileTransfer *transfer,
-                        const char *path,
-                        struct stat *st_buf,
-                        struct stat *parent_st_buf)
+                        const char   *path,
+                        struct stat  *st_buf,
+                        struct stat  *parent_st_buf)
 {
   ExportedFile *file;
 
@@ -243,9 +249,9 @@ file_transfer_add_file (FileTransfer *transfer,
 }
 
 static char **
-file_transfer_execute (FileTransfer *transfer,
-                       XdpAppInfo *target_app_info,
-                       GError **error)
+file_transfer_execute (FileTransfer  *transfer,
+                       XdpAppInfo    *target_app_info,
+                       GError       **error)
 {
   DocumentAddFullFlags common_flags;
   DocumentPermissionFlags perms;
@@ -338,8 +344,8 @@ file_transfer_execute (FileTransfer *transfer,
 
 static void
 start_transfer (GDBusMethodInvocation *invocation,
-                GVariant *parameters,
-                XdpAppInfo *app_info)
+                GVariant              *parameters,
+                XdpAppInfo            *app_info)
 {
   g_autoptr(GVariant) options = NULL;
   g_autoptr(FileTransfer) transfer = NULL;
@@ -363,8 +369,8 @@ start_transfer (GDBusMethodInvocation *invocation,
 
 static void
 add_files (GDBusMethodInvocation *invocation,
-           GVariant *parameters,
-           XdpAppInfo *app_info)
+           GVariant              *parameters,
+           XdpAppInfo            *app_info)
 {
   FileTransfer *transfer;
   const char *key;
@@ -456,8 +462,8 @@ add_files (GDBusMethodInvocation *invocation,
 
 static void
 retrieve_files (GDBusMethodInvocation *invocation,
-                GVariant *parameters,
-                XdpAppInfo *app_info)
+                GVariant              *parameters,
+                XdpAppInfo            *app_info)
 {
   const char *key;
   FileTransfer *transfer;
@@ -490,8 +496,8 @@ retrieve_files (GDBusMethodInvocation *invocation,
 
 static void
 stop_transfer (GDBusMethodInvocation *invocation,
-                GVariant *parameters,
-                XdpAppInfo *app_info)
+               GVariant              *parameters,
+               XdpAppInfo            *app_info)
 {
   const char *key;
   FileTransfer *transfer;
@@ -527,7 +533,10 @@ handle_method (GCallback              method_callback,
   g_autoptr(XdpAppInfo) app_info = NULL;
   PortalMethod portal_method = (PortalMethod)method_callback;
 
-  app_info = xdp_invocation_ensure_app_info_sync (invocation, NULL, &error);
+  app_info = xdp_app_info_registry_ensure_for_invocation_sync (app_info_registry,
+                                                               invocation,
+                                                               NULL,
+                                                               &error);
   if (app_info == NULL)
     g_dbus_method_invocation_return_gerror (invocation, error);
   else
@@ -553,7 +562,7 @@ file_transfer_create (void)
   return G_DBUS_INTERFACE_SKELETON (file_transfer);
 }
 
-void
+static void
 stop_file_transfers_in_thread_func (GTask        *task,
                                     gpointer      source_object,
                                     gpointer      task_data,
@@ -583,10 +592,9 @@ stop_file_transfers_in_thread_func (GTask        *task,
 void
 stop_file_transfers_for_sender (const char *sender)
 {
-  GTask *task;
+  g_autoptr(GTask) task = NULL;
 
   task = g_task_new (NULL, NULL, NULL, NULL);
   g_task_set_task_data (task, g_strdup (sender), g_free);
   g_task_run_in_thread (task, stop_file_transfers_in_thread_func);
-  g_object_unref (task);
 }
