@@ -20,23 +20,21 @@
 
 #include "config.h"
 
+#include <stdio.h>
 #include <glib/gi18n.h>
 #include <gio/gunixfdlist.h>
 #include <gio/gdesktopappinfo.h>
-#include <stdio.h>
 
-#include "xdp-request.h"
-#include "xdp-permissions.h"
 #include "pipewire.h"
+#include "xdp-context.h"
 #include "xdp-dbus.h"
 #include "xdp-impl-dbus.h"
+#include "xdp-permissions.h"
+#include "xdp-portal-config.h"
+#include "xdp-request.h"
 #include "xdp-utils.h"
 
-#define PERMISSION_TABLE "devices"
-#define PERMISSION_DEVICE_CAMERA "camera"
-
-static XdpDbusImplLockdown *lockdown;
-static XdpDbusImplAccess *access_impl;
+#include "camera.h"
 
 typedef struct _Camera Camera;
 typedef struct _CameraClass CameraClass;
@@ -45,6 +43,9 @@ struct _Camera
 {
   XdpDbusCameraSkeleton parent_instance;
 
+  XdpDbusImplAccess *access_impl;
+  XdpDbusImplLockdown *lockdown_impl;
+
   PipeWireRemote *pipewire_remote;
   GSource *pipewire_source;
   GFileMonitor *pipewire_socket_monitor;
@@ -58,8 +59,6 @@ struct _CameraClass
   XdpDbusCameraSkeletonClass parent_class;
 };
 
-static Camera *camera;
-
 GType camera_get_type (void);
 static void camera_iface_init (XdpDbusCameraIface *iface);
 
@@ -67,55 +66,59 @@ G_DEFINE_TYPE_WITH_CODE (Camera, camera, XDP_DBUS_TYPE_CAMERA_SKELETON,
                          G_IMPLEMENT_INTERFACE (XDP_DBUS_TYPE_CAMERA,
                                                 camera_iface_init))
 
+G_DEFINE_AUTOPTR_CLEANUP_FUNC (Camera, g_object_unref)
+
 static gboolean
 create_pipewire_remote (Camera *camera,
                         GError **error);
 
 static gboolean
-query_permission_sync (XdpRequest *request)
+query_permission_sync (Camera     *camera,
+                       XdpRequest *request)
 {
   XdpPermission permission;
-  const char *app_id;
+  XdpAppInfo *app_info = request->app_info;
   gboolean allowed;
 
-  app_id = (const char *)g_object_get_data (G_OBJECT (request), "app-id");
-  permission = xdp_get_permission_sync (app_id, PERMISSION_TABLE, PERMISSION_DEVICE_CAMERA);
+  permission = xdp_get_permission_sync (app_info,
+                                        CAMERA_PERMISSION_TABLE,
+                                        CAMERA_PERMISSION_DEVICE_CAMERA);
   if (permission == XDP_PERMISSION_ASK || permission == XDP_PERMISSION_UNSET)
     {
       g_auto(GVariantBuilder) opt_builder =
         G_VARIANT_BUILDER_INIT (G_VARIANT_TYPE_VARDICT);
       g_autofree char *title = NULL;
-      g_autofree char *body = NULL;
+      g_autofree char *subtitle = NULL;
+      const char *body;
       guint32 response = 2;
       g_autoptr(GVariant) results = NULL;
       g_autoptr(GError) error = NULL;
-      g_autoptr(GAppInfo) info = NULL;
       g_autoptr(XdpDbusImplRequest) impl_request = NULL;
-
-      if (app_id[0] != 0)
-        {
-          g_autofree char *desktop_id = g_strconcat (app_id, ".desktop", NULL);
-          info = (GAppInfo*)g_desktop_app_info_new (desktop_id);
-        }
+      const char *app_id = xdp_app_info_get_id (app_info);
+      const char *app_name = xdp_app_info_get_app_display_name (app_info);
 
       g_variant_builder_add (&opt_builder, "{sv}", "icon", g_variant_new_string ("camera-web-symbolic"));
 
-      if (info)
+      if (app_name)
         {
-          title = g_strdup_printf (_("Allow %s to Use the Camera?"), g_app_info_get_display_name (info));
-          body = g_strdup_printf (_("%s wants to access camera devices."), g_app_info_get_display_name (info));
+          title = g_strdup_printf (_("Allow %s to Use the Camera?"), app_name);
+          subtitle = g_strdup_printf (_("%s wants to access the camera"), app_name);
         }
       else
         {
-          title = g_strdup (_("Allow app to Use the Camera?"));
-          body = g_strdup (_("An app wants to access camera devices."));
+          title = g_strdup (_("Allow Apps to Use the Camera?"));
+          subtitle = g_strdup (_("An app wants to access the camera"));
         }
 
-      impl_request = xdp_dbus_impl_request_proxy_new_sync (g_dbus_proxy_get_connection (G_DBUS_PROXY (access_impl)),
-                                                           G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
-                                                           g_dbus_proxy_get_name (G_DBUS_PROXY (access_impl)),
-                                                           request->id,
-                                                           NULL, &error);
+      body = _("This permission can be changed at any time from the privacy settings");
+
+      impl_request = xdp_dbus_impl_request_proxy_new_sync (
+        g_dbus_proxy_get_connection (G_DBUS_PROXY (camera->access_impl)),
+        G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
+        g_dbus_proxy_get_name (G_DBUS_PROXY (camera->access_impl)),
+        request->id,
+        NULL, &error);
+
       if (!impl_request)
         return FALSE;
 
@@ -123,12 +126,12 @@ query_permission_sync (XdpRequest *request)
 
       g_debug ("Calling backend for device access to camera");
 
-      if (!xdp_dbus_impl_access_call_access_dialog_sync (access_impl,
+      if (!xdp_dbus_impl_access_call_access_dialog_sync (camera->access_impl,
                                                          request->id,
                                                          app_id,
                                                          "",
                                                          title,
-                                                         "",
+                                                         subtitle,
                                                          body,
                                                          g_variant_builder_end (&opt_builder),
                                                          &response,
@@ -147,8 +150,10 @@ query_permission_sync (XdpRequest *request)
 
       if (permission == XDP_PERMISSION_UNSET)
         {
-          xdp_set_permission_sync (app_id, PERMISSION_TABLE, PERMISSION_DEVICE_CAMERA,
-                               allowed ? XDP_PERMISSION_YES : XDP_PERMISSION_NO);
+          xdp_set_permission_sync (app_info,
+                                   CAMERA_PERMISSION_TABLE,
+                                   CAMERA_PERMISSION_DEVICE_CAMERA,
+                                   allowed ? XDP_PERMISSION_YES : XDP_PERMISSION_NO);
         }
     }
   else
@@ -163,10 +168,11 @@ handle_access_camera_in_thread_func (GTask *task,
                                      gpointer task_data,
                                      GCancellable *cancellable)
 {
+  Camera *camera = (Camera *) source_object;
   XdpRequest *request = XDP_REQUEST (task_data);
   gboolean allowed;
 
-  allowed = query_permission_sync (request);
+  allowed = query_permission_sync (camera, request);
 
   REQUEST_AUTOLOCK (request);
 
@@ -186,26 +192,16 @@ handle_access_camera_in_thread_func (GTask *task,
     }
 }
 
-static const char *
-app_id_from_app_info (XdpAppInfo *app_info)
-{
-  /* Automatically grant camera access to unsandboxed apps. */
-  if (xdp_app_info_is_host (app_info))
-    return "";
-
-  return xdp_app_info_get_id (app_info);
-}
-
 static gboolean
 handle_access_camera (XdpDbusCamera *object,
                       GDBusMethodInvocation *invocation,
                       GVariant *arg_options)
 {
+  Camera *camera = (Camera *) object;
   XdpRequest *request = xdp_request_from_invocation (invocation);
-  const char *app_id;
   g_autoptr(GTask) task = NULL;
 
-  if (xdp_dbus_impl_lockdown_get_disable_camera (lockdown))
+  if (xdp_dbus_impl_lockdown_get_disable_camera (camera->lockdown_impl))
     {
       g_debug ("Camera access disabled");
       g_dbus_method_invocation_return_error (invocation,
@@ -217,14 +213,11 @@ handle_access_camera (XdpDbusCamera *object,
 
   REQUEST_AUTOLOCK (request);
 
-  app_id = app_id_from_app_info (request->app_info);
-  g_object_set_data_full (G_OBJECT (request), "app-id", g_strdup (app_id), g_free);
-
   xdp_request_export (request, g_dbus_method_invocation_get_connection (invocation));
 
   xdp_dbus_camera_complete_access_camera (object, invocation, request->id);
 
-  task = g_task_new (object, NULL, NULL, NULL);
+  task = g_task_new (camera, NULL, NULL, NULL);
   g_task_set_task_data (task, g_object_ref (request), g_object_unref);
   g_task_run_in_thread (task, handle_access_camera_in_thread_func);
 
@@ -272,8 +265,8 @@ handle_open_pipewire_remote (XdpDbusCamera *object,
                              GUnixFDList *in_fd_list,
                              GVariant *arg_options)
 {
+  Camera *camera = (Camera *) object;
   g_autoptr(XdpAppInfo) app_info = NULL;
-  const char *app_id;
   XdpPermission permission;
   g_autoptr(GUnixFDList) out_fd_list = NULL;
   int fd;
@@ -281,7 +274,7 @@ handle_open_pipewire_remote (XdpDbusCamera *object,
   g_autoptr(GError) error = NULL;
   PipeWireRemote *remote;
 
-  if (xdp_dbus_impl_lockdown_get_disable_camera (lockdown))
+  if (xdp_dbus_impl_lockdown_get_disable_camera (camera->lockdown_impl))
     {
       g_debug ("Camera access disabled");
       g_dbus_method_invocation_return_error (invocation,
@@ -291,9 +284,10 @@ handle_open_pipewire_remote (XdpDbusCamera *object,
       return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
-  app_info = xdp_invocation_ensure_app_info_sync (invocation, NULL, &error);
-  app_id = app_id_from_app_info (app_info);
-  permission = xdp_get_permission_sync (app_id, PERMISSION_TABLE, PERMISSION_DEVICE_CAMERA);
+  app_info = xdp_invocation_get_app_info (invocation);
+  permission = xdp_get_permission_sync (app_info,
+                                        CAMERA_PERMISSION_TABLE,
+                                        CAMERA_PERMISSION_DEVICE_CAMERA);
   if (permission != XDP_PERMISSION_YES)
     {
       g_dbus_method_invocation_return_error (invocation,
@@ -303,7 +297,7 @@ handle_open_pipewire_remote (XdpDbusCamera *object,
       return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
-  remote = open_pipewire_camera_remote (app_id, &error);
+  remote = open_pipewire_camera_remote (xdp_app_info_get_id (app_info), &error);
   if (!remote)
     {
       g_dbus_method_invocation_return_error (invocation,
@@ -478,40 +472,15 @@ on_pipewire_socket_changed (GFileMonitor *monitor,
     g_warning ("Failed connect to PipeWire: %s", error->message);
 }
 
-static gboolean
-init_camera_tracker (Camera *camera,
-                     GError **error)
-{
-  g_autofree char *pipewire_socket_path = NULL;
-  g_autoptr(GFile) pipewire_socket = NULL;
-  g_autoptr(GError) local_error = NULL;
-
-  pipewire_socket_path = g_strdup_printf ("%s/pipewire-0",
-                                          g_get_user_runtime_dir ());
-  pipewire_socket = g_file_new_for_path (pipewire_socket_path);
-  camera->pipewire_socket_monitor =
-    g_file_monitor_file (pipewire_socket, G_FILE_MONITOR_NONE, NULL, error);
-  if (!camera->pipewire_socket_monitor)
-    return FALSE;
-
-  g_signal_connect (camera->pipewire_socket_monitor,
-                    "changed",
-                    G_CALLBACK (on_pipewire_socket_changed),
-                    camera);
-
-  camera->cameras = g_hash_table_new (NULL, NULL);
-
-  if (!create_pipewire_remote (camera, &local_error))
-    g_warning ("Failed connect to PipeWire: %s", local_error->message);
-
-  return TRUE;
-}
-
 static void
 camera_finalize (GObject *object)
 {
   Camera *camera = (Camera *)object;
 
+  g_clear_object (&camera->access_impl);
+  g_clear_object (&camera->lockdown_impl);
+  g_clear_object (&camera->pipewire_socket_monitor);
+
   g_clear_pointer (&camera->pipewire_source, g_source_destroy);
   g_clear_pointer (&camera->pipewire_remote, pipewire_remote_destroy);
   g_clear_pointer (&camera->cameras, g_hash_table_unref);
@@ -522,12 +491,6 @@ camera_finalize (GObject *object)
 static void
 camera_init (Camera *camera)
 {
-  g_autoptr(GError) error = NULL;
-
-  xdp_dbus_camera_set_version (XDP_DBUS_CAMERA (camera), 1);
-
-  if (!init_camera_tracker (camera, &error))
-    g_warning ("Failed to track cameras: %s", error->message);
 }
 
 static void
@@ -538,30 +501,65 @@ camera_class_init (CameraClass *klass)
   object_class->finalize = camera_finalize;
 }
 
-GDBusInterfaceSkeleton *
-camera_create (GDBusConnection *connection,
-               const char      *access_impl_dbus_name,
-               gpointer         lockdown_proxy)
+static Camera *
+camera_new (XdpDbusImplAccess   *access_impl,
+            XdpDbusImplLockdown *lockdown_impl)
 {
+  Camera *camera;
+  g_autofree char *pipewire_socket_path = NULL;
+  g_autoptr(GFile) pipewire_socket = NULL;
   g_autoptr(GError) error = NULL;
 
-  lockdown = lockdown_proxy;
-
   camera = g_object_new (camera_get_type (), NULL);
+  camera->access_impl = g_object_ref (access_impl);
+  camera->lockdown_impl = g_object_ref (lockdown_impl);
 
-  access_impl = xdp_dbus_impl_access_proxy_new_sync (connection,
-                                                     G_DBUS_PROXY_FLAGS_NONE,
-                                                     access_impl_dbus_name,
-                                                     DESKTOP_PORTAL_OBJECT_PATH,
-                                                     NULL,
-                                                     &error);
-  if (access_impl == NULL)
+  xdp_dbus_camera_set_version (XDP_DBUS_CAMERA (camera), 1);
+
+  pipewire_socket_path = g_strdup_printf ("%s/pipewire-0",
+                                          g_get_user_runtime_dir ());
+  pipewire_socket = g_file_new_for_path (pipewire_socket_path);
+  camera->pipewire_socket_monitor =
+    g_file_monitor_file (pipewire_socket, G_FILE_MONITOR_NONE, NULL, &error);
+  if (!camera->pipewire_socket_monitor)
     {
-      g_warning ("Failed to create access proxy: %s", error->message);
-      return NULL;
+      g_warning ("Failed to track cameras: %s", error->message);
+    }
+  else
+    {
+      g_signal_connect_object (camera->pipewire_socket_monitor, "changed",
+                               G_CALLBACK (on_pipewire_socket_changed),
+                               camera,
+                               G_CONNECT_DEFAULT);
+
+      camera->cameras = g_hash_table_new (NULL, NULL);
+
+      if (!create_pipewire_remote (camera, &error))
+        g_warning ("Failed to connect to PipeWire: %s", error->message);
     }
 
-  g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (access_impl), G_MAXINT);
+  return camera;
+}
 
-  return G_DBUS_INTERFACE_SKELETON (camera);
+void
+init_camera (XdpContext *context)
+{
+  g_autoptr(Camera) camera = NULL;
+  XdpDbusImplAccess *access_impl;
+  XdpDbusImplLockdown *lockdown_impl;
+
+  access_impl = xdp_context_get_access_impl (context);
+  if (access_impl == NULL)
+    {
+      g_warning ("The camera portal requires an access impl");
+      return;
+    }
+
+  lockdown_impl = xdp_context_get_lockdown_impl (context);
+
+  camera = camera_new (access_impl, lockdown_impl);
+
+  xdp_context_take_and_export_portal (context,
+                                      G_DBUS_INTERFACE_SKELETON (g_steal_pointer (&camera)),
+                                      XDP_CONTEXT_EXPORT_FLAGS_NONE);
 }
