@@ -22,6 +22,6 @@
 
 #pragma once
 
-#include <gio/gio.h>
+#include "xdp-types.h"
 
-GDBusInterfaceSkeleton * game_mode_create (GDBusConnection *connection);
+void init_game_mode (XdpContext *context);
