@@ -20,8 +20,6 @@
 
 #pragma once
 
-#include <gio/gio.h>
-#include <stdint.h>
+#include "xdp-types.h"
 
-GDBusInterfaceSkeleton * input_capture_create (GDBusConnection *connection,
-                                               const char      *dbus_name);
+void init_input_capture (XdpContext *context);
