@@ -23,8 +23,6 @@
 
 #pragma once
 
-#include <gio/gio.h>
+#include "xdp-types.h"
 
-GDBusInterfaceSkeleton * open_uri_create (GDBusConnection *connection,
-                                          const char      *dbus_name,
-                                          gpointer         lockdown);
+void init_open_uri (XdpContext *context);
