@@ -22,7 +22,7 @@
 
 #pragma once
 
-#include <gio/gio.h>
+#include "xdp-types.h"
 
-GDBusInterfaceSkeleton * secret_create (GDBusConnection *connection,
-					const char *dbus_name);
+void init_secret (XdpContext   *context,
+                  GCancellable *cancellable);
