@@ -32,15 +32,18 @@ G_DECLARE_FINAL_TYPE (XdpAppInfoHost,
                       XDP, APP_INFO_HOST,
                       XdpAppInfo)
 
-XdpAppInfo * xdp_app_info_host_new (int pid,
-                                    int pidfd);
+XdpAppInfo * xdp_app_info_host_new (const char *sender,
+                                    int         pid,
+                                    int        *pidfd);
 
 XdpAppInfo *
-xdp_app_info_host_new_registered (int          pidfd,
+xdp_app_info_host_new_registered (const char  *sender,
+                                  int          pid,
+                                  int          pidfd,
                                   const char  *app_id,
                                   GError     **error);
 
-#ifdef HAVE_LIBSYSTEMD
+#if HAVE_LIBSYSTEMD
 XDP_EXPORT_TEST
 char * _xdp_app_info_host_parse_app_id_from_unit_name (const char *unit);
 #endif
