@@ -2,7 +2,9 @@
 #
 # This file is formatted with Python Black
 
-from typing import Any, Dict, Iterator, Optional
+import tests.xdp_utils as xdp
+
+from typing import Any, Dict, Iterator, Optional, assert_never
 from types import ModuleType
 
 import pytest
@@ -393,31 +395,67 @@ def xdp_overwrite_env() -> dict[str, str]:
     return {}
 
 
-@pytest.fixture
-def app_id() -> str:
+@pytest.fixture(
+    params=[
+        xdp.AppInfoKind.HOST,
+        xdp.AppInfoKind.FLATPAK,
+        xdp.AppInfoKind.SNAP,
+        xdp.AppInfoKind.LINYAPS,
+    ]
+)
+def xdp_app_info(request) -> xdp.AppInfo:
     """
-    Default fixture which can be used to override the app id that the portal
-    frontend will discover for incoming connections.
+    Default fixture which can be used to override the XdpAppInfo the portal
+    frontend will discover.
+    The default fixture is parametric and will run each test with all the
+    app info kinds.
     """
-    return "org.example.Test"
+
+    app_info_kind = request.param
+    app_id = "org.example.Test"
+
+    if app_info_kind == xdp.AppInfoKind.HOST:
+        return xdp.AppInfo.new_host(
+            app_id=app_id,
+        )
+
+    if app_info_kind == xdp.AppInfoKind.FLATPAK:
+        return xdp.AppInfo.new_flatpak(
+            app_id=app_id,
+        )
+
+    if app_info_kind == xdp.AppInfoKind.SNAP:
+        return xdp.AppInfo.new_snap(
+            common_id=app_id,
+            snap_name="test",
+            app_name="test",
+        )
+
+    if app_info_kind == xdp.AppInfoKind.LINYAPS:
+        return xdp.AppInfo.new_linyaps(
+            app_id=app_id,
+        )
+
+    assert_never(app_info_kind)
+
+
+@pytest.fixture(autouse=True)
+def ensure_xdp_app_info_files(create_test_dirs, xdp_app_info) -> None:
+    xdp_app_info.ensure_files()
 
 
 @pytest.fixture
 def xdp_env(
     xdp_overwrite_env: dict[str, str],
-    app_id: str,
-    usb_queries: Optional[str],
+    xdp_app_info: xdp.AppInfo,
     umockdev: Optional[UMockdev.Testbed],
 ) -> dict[str, str]:
     env = os.environ.copy()
     env["G_DEBUG"] = "fatal-criticals"
     env["XDG_CURRENT_DESKTOP"] = "test"
 
-    if app_id:
-        env["XDG_DESKTOP_PORTAL_TEST_APP_ID"] = app_id
-
-    if usb_queries:
-        env["XDG_DESKTOP_PORTAL_TEST_USB_QUERIES"] = usb_queries
+    if xdp_app_info:
+        xdp_app_info.extend_env(env)
 
     if umockdev:
         env["UMOCKDEV_DIR"] = umockdev.get_root_dir()
@@ -450,9 +488,45 @@ def _maybe_add_asan_preload(executable: Path, env: dict[str, str]) -> None:
     env["LD_PRELOAD"] = f"{libasan}:{preload}"
 
 
+def _valgrind_glib_suppressions() -> str:
+    try:
+        import pkgconfig
+
+        return pkgconfig.variables("glib-2.0")["glib_valgrind_suppressions"]
+    except ImportError:
+        return "/usr/share/glib-2.0/valgrind/glib.supp"
+
+
+@pytest.fixture
+def xdp_valgrind_args() -> list[str]:
+    if not xdp.is_valgrind():
+        return []
+
+    xdp_suppression = test_dir() / "valgrind.suppression"
+    glib_suppression = _valgrind_glib_suppressions()
+
+    return [
+        "valgrind",
+        "--tool=memcheck",
+        "--error-exitcode=1",
+        "--track-origins=yes",
+        "--leak-check=full",
+        "--leak-resolution=high",
+        "--num-callers=50",
+        "--show-leak-kinds=definite,possible",
+        "--show-error-list=yes",
+        "--show-realloc-size-zero=no",
+        f"--suppressions={glib_suppression}",
+        f"--suppressions={xdp_suppression}",
+    ]
+
+
 @pytest.fixture
 def xdg_desktop_portal(
-    dbus_con: dbus.Bus, xdg_desktop_portal_path: Path, xdp_env: dict[str, str]
+    dbus_con: dbus.Bus,
+    xdg_desktop_portal_path: Path,
+    xdp_env: dict[str, str],
+    xdp_valgrind_args: list[str],
 ) -> Iterator[subprocess.Popen]:
     """
     Fixture which starts and eventually stops xdg-desktop-portal
@@ -463,7 +537,9 @@ def xdg_desktop_portal(
     env = xdp_env.copy()
     _maybe_add_asan_preload(xdg_desktop_portal_path, env)
 
-    xdg_desktop_portal = subprocess.Popen([xdg_desktop_portal_path], env=env)
+    xdg_desktop_portal = subprocess.Popen(
+        xdp_valgrind_args + [xdg_desktop_portal_path], env=env
+    )
 
     while not dbus_con.name_has_owner("org.freedesktop.portal.Desktop"):
         returncode = xdg_desktop_portal.poll()
@@ -527,6 +603,12 @@ def xdg_document_portal(
     env = xdp_env.copy()
     env.pop("LD_PRELOAD", None)
 
+    # The document portal also uses XdpAppInfos to figure out where a request
+    # is coming from. If we don't unset this, it might think the
+    # xdg-desktop-portal is a flatpak or snap app. Would be nice to find a
+    # better solution here.
+    env.pop("XDG_DESKTOP_PORTAL_TEST_APP_INFO_KIND", None)
+
     document_portal = subprocess.Popen([xdg_document_portal_path], env=env)
 
     while not dbus_con.name_has_owner("org.freedesktop.portal.Documents"):
@@ -543,6 +625,17 @@ def xdg_document_portal(
     returncode = document_portal.wait()
     assert returncode == 0
 
+    fuse_mount = Path(os.environ["XDG_RUNTIME_DIR"]) / "doc"
+
+    def unmounted():
+        try:
+            next(fuse_mount.iterdir())
+        except StopIteration:
+            return True
+        return False
+
+    xdp.wait_for(unmounted)
+
 
 @pytest.fixture
 def portals(templates: Any, xdg_desktop_portal: Any, xdg_permission_store: Any) -> None:
@@ -553,15 +646,6 @@ def portals(templates: Any, xdg_desktop_portal: Any, xdg_permission_store: Any)
     return None
 
 
-@pytest.fixture
-def usb_queries() -> Optional[str]:
-    """
-    Default fixture providing the usb queries the connecting process can
-    enumerate
-    """
-    return None
-
-
 @pytest.fixture
 def umockdev() -> Optional[UMockdev.Testbed]:
     """
